import { YOUTUBE_CLIENT_ID, YOUTUBE_CLIENT_SECRET, YOUTUBE_PLAYLIST_ID } from '$env/static/private';
import { Logger } from '$lib/log';
import type { OauthResponse } from '$lib/mastodon/response';
import type {
  YoutubePlaylistItemResponse,
  YoutubeResponse
} from '$lib/server/playlist/youtubeResponse';
import type { SongInfo } from '$lib/odesliResponse';
import { OauthPlaylistAdder } from './oauthPlaylistAdder';
import type { PlaylistAdder } from './playlistAdder';

export class YoutubePlaylistAdder extends OauthPlaylistAdder implements PlaylistAdder {
  public constructor() {
    super('https://www.googleapis.com/youtube/v3', 'yt_auth_token');
    this.logger = new Logger('YoutubePlaylistAdder');
  }

  public constructAuthUrl(redirectUri: URL): URL {
    this.redirectUri = redirectUri;
    let additionalParameters = new Map([
      ['access_type', 'offline'],
      ['include_granted_scopes', 'false']
    ]);
    const endpoint = 'https://accounts.google.com/o/oauth2/v2/auth';
    return this.constructAuthUrlInternal(
      endpoint,
      YOUTUBE_CLIENT_ID,
      'https://www.googleapis.com/auth/youtube',
      redirectUri,
      additionalParameters
    );
  }

  public async receivedAuthCode(code: string, url: URL) {
    this.logger.debug('received code');
    this.redirectUri = url;
    const tokenUrl = new URL('https://oauth2.googleapis.com/token');
    await this.receivedAuthCodeInternal(
      tokenUrl,
      YOUTUBE_CLIENT_ID,
      code,
      url,
      YOUTUBE_CLIENT_SECRET
    );
    const token = await this.refreshToken();
    if (token == null) {
      return;
    }
  }

  private async checkAuthorizedToken(token: OauthResponse) {
    try {
      this.logger.debug('Checking authorized token');
      const res = await fetch(this.apiBase + '/channels?part=id&mine=true', {
        method: 'GET',
        headers: { Authorization: `${token.token_type} ${token.access_token}` }
      }).then((r) => r.json());
      this.logger.debug('Checked authorized token', res);
    } catch (e) {
      this.logger.error('Error checking authorized token', e);
    }
  }

  private async refreshToken(force: boolean = false): Promise<OauthResponse | null> {
    const tokenInfo = await this.shouldRefreshToken();
    if (tokenInfo == null) {
      return null;
    }
    let token = tokenInfo.token;
    if (!tokenInfo.refresh && !force) {
      return token;
    }
    if (!token.refresh_token) {
      this.logger.error('Need to refresh access token, but no refresh token provided');
      return null;
    }

    const tokenUrl = new URL('https://oauth2.googleapis.com/token');
    const refreshedToken = await this.requestRefreshToken(
      tokenUrl,
      YOUTUBE_CLIENT_ID,
      token.refresh_token,
      this.getRedirectUri('ytauth').toString(),
      YOUTUBE_CLIENT_SECRET
    );
    if (refreshedToken !== null) {
      await this.checkAuthorizedToken(refreshedToken);
    }
    return refreshedToken;
  }

  public async addToPlaylist(song: SongInfo) {
    this.logger.debug('addToYoutubePlaylist');
    let token = await this.refreshToken();
    if (token == null) {
      return;
    }

    if (!YOUTUBE_PLAYLIST_ID || YOUTUBE_PLAYLIST_ID === 'CHANGE_ME') {
      this.logger.debug('no playlist ID configured');
      return;
    }
    if (!song.youtubeUrl) {
      this.logger.info('Skip adding song to YT playlist, no youtube Url', song);
      return;
    }

    const songUrl = new URL(song.youtubeUrl);
    let videoIds: string[] = [];
    if (song.type === 'album') {
      const albumPlaylistId = songUrl.searchParams.get('list') ?? '';
      const albumItemsUrl = new URL(this.apiBase + '/playlistItems');
      albumItemsUrl.searchParams.append('maxResults', '50');
      albumItemsUrl.searchParams.append('playlistId', albumPlaylistId);
      albumItemsUrl.searchParams.append('part', 'snippet');
      const albumPlaylistItem = await fetch(albumItemsUrl, {
        headers: { Authorization: `${token.token_type} ${token.access_token}` }
      }).then((r) => r.json());
      const albumTracks: any[] = albumPlaylistItem.items ?? [];
      videoIds = albumTracks.map((x) => x.snippet?.resourceId?.videoId).filter((x) => x);
      this.logger.info(
        'Found',
        albumPlaylistItem.pageInfo?.totalResults,
        'songs in album, received',
        albumTracks.length
      );
      this.logger.debug(videoIds);
      if (videoIds.length === 0) {
        this.logger.debug(
          'Skip adding album to YT playlist, is empty',
          song.youtubeUrl,
          albumPlaylistItem
        );
        return;
      }
    } else {
      const youtubeId = songUrl.searchParams.get('v');
      if (!youtubeId) {
        this.logger.debug(
          'Skip adding song to YT playlist, could not extract YT id from URL',
          song.youtubeUrl
        );
        return;
      }
      this.logger.debug('Found YT id from URL', song.youtubeUrl, youtubeId);
      videoIds.push(youtubeId);
    }

    for (let youtubeId of videoIds) {
      this.logger.debug('Adding to playlist', youtubeId);
      const alreadyAdded = await this.isVideoInPlaylist(youtubeId, token);
      if (alreadyAdded) {
        this.logger.info('Item already in playlist', song.youtubeUrl, song.title);
        continue;
      } else {
        this.logger.debug('Item not already in playlist', song.youtubeUrl, song.title);
      }

      let retries = 3;
      let success = false;
      while (retries > 0 && !success) {
        try {
          this.logger.debug('Retries', retries);
          await this.addSongToPlaylist(youtubeId, token);
          success = true;
          this.logger.info('Added to playlist', youtubeId, song.title);
        } catch (e) {
          retries--;
          if (e instanceof Error && e.message === 'authError') {
            this.logger.info('Refreshing auth token');
            token = await this.refreshToken(true);
            if (token == null) {
              this.logger.error('Refreshing auth token failed');
              return;
            }
          } else {
            this.logger.error('Add to playlist failed', e);
          }
        }
      }
    }
  }

  private async addSongToPlaylist(videoId: string, token: OauthResponse): Promise<void> {
    const addItemUrl = new URL(this.apiBase + '/playlistItems');
    addItemUrl.searchParams.append('part', 'snippet');
    const options: RequestInit = {
      method: 'POST',
      headers: { Authorization: `${token.token_type} ${token.access_token}` },
      body: JSON.stringify({
        snippet: {
          playlistId: YOUTUBE_PLAYLIST_ID,
          resourceId: {
            videoId: videoId,
            kind: 'youtube#video'
          }
        }
      })
    };
    const request = new Request(addItemUrl, options);
    const resp = await fetch(request);
    let respObj: YoutubeResponse = await resp.json();
    if (respObj.error) {
      this.logger.error(
        'Add to playlist failed',
        respObj.error.errors,
        respObj.error.code,
        respObj.error.message
      );
      throw new Error(respObj.error.errors[0].reason);
    }
  }

  private async isVideoInPlaylist(videoId: string, token: OauthResponse): Promise<boolean> {
    const playlistItemsUrl = new URL(this.apiBase + '/playlistItems');
    playlistItemsUrl.searchParams.append('videoId', videoId);
    playlistItemsUrl.searchParams.append('playlistId', YOUTUBE_PLAYLIST_ID);
    playlistItemsUrl.searchParams.append('part', 'id');
    const existingPlaylistItem: YoutubePlaylistItemResponse = await fetch(playlistItemsUrl, {
      headers: { Authorization: `${token.token_type} ${token.access_token}` }
    }).then((r) => r.json());
    if (existingPlaylistItem.error) {
      this.logger.error(
        'Could not check if item is already in playlist',
        existingPlaylistItem.error
      );
    }
    return existingPlaylistItem.pageInfo && existingPlaylistItem.pageInfo.totalResults > 0;
  }
}
